# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb-ctl commands and command handlers specific to the repository"""
import os

# *ctl module should limit the number of import to be imported as quickly as
# possible (for cubicweb-ctl reactivity, necessary for instance for usable bash
# completion). So import locally in command helpers.
import sched
import sys
from contextlib import contextmanager

from glob import glob
from tempfile import NamedTemporaryFile

from logilab.common.configuration import Configuration, merge_options
from logilab.common.shellutils import ASK, generate_password
from logilab.database import get_db_helper, get_connection

from cubicweb import (
    AuthenticationError,
    ExecutionError,
    ConfigurationError,
)
from cubicweb.cwctl import CWCTL, check_options_consistency, ConfigureInstanceCommand
from cubicweb.server import SOURCE_TYPES
from cubicweb.server import checkintegrity
from cubicweb.server.check_unused_index import check_unused_index
from cubicweb.server.serverconfig import (
    USER_OPTIONS,
    SourceConfiguration,
    ask_source_config,
    generate_source_config,
)
from cubicweb.cwconfig import CubicWebConfiguration
from cubicweb.toolsutils import Command, CommandHandler, underline_title


# utility functions ###########################################################


def source_cnx(source, dbname=None, special_privs=False, interactive=True):
    """open and return a connection to the system database defined in the
    given server.serverconfig
    """
    from getpass import getpass

    dbhost = source.get("db-host")
    if dbname is None:
        dbname = source["db-name"]
    driver = source["db-driver"]
    dbhelper = get_db_helper(driver)
    if interactive:
        print(f"-> connecting to {driver} database", end=" ")
        if dbhost:
            print(f"{dbname}@{dbhost}", end=" ")
        else:
            print(dbname, end=" ")
    if dbhelper.users_support:
        if not interactive or (not special_privs and source.get("db-user")):
            user = source.get("db-user", os.environ.get("USER", ""))
            if interactive:
                print("as", user)
            password = source.get("db-password")
        else:
            print()
            if special_privs:
                print("WARNING")
                print(
                    "the user will need the following special access rights "
                    "on the database:"
                )
                print(special_privs)
                print()
            default_user = source.get("db-user", os.environ.get("USER", ""))
            user = input(f"Connect as user ? [{default_user!r}]: ")
            user = user.strip() or default_user
            if user == source.get("db-user"):
                password = source.get("db-password")
            else:
                password = getpass("password: ")
    else:
        user = password = None
    extra_args = source.get("db-extra-arguments")
    extra = extra_args and {"extra_args": extra_args} or {}
    cnx = get_connection(
        driver,
        dbhost,
        dbname,
        user,
        password=password,
        port=source.get("db-port"),
        schema=source.get("db-namespace"),
        **extra,
    )
    try:
        cnx.logged_user = user
    except AttributeError:
        # C object, __slots__
        from logilab.database import _SimpleConnectionWrapper

        cnx = _SimpleConnectionWrapper(cnx)
        cnx.logged_user = user
    return cnx


def _db_sys_cnx(source, special_privs, interactive=True):
    """return a connection on the RDMS system table (to create/drop a user or a
    database)
    """
    import logilab.common as lgp

    lgp.USE_MX_DATETIME = False
    # connect on the dbms system base to create our base
    system_db = get_db_helper(source["db-driver"]).system_database()
    cnx = source_cnx(
        source, system_db, special_privs=special_privs, interactive=interactive
    )
    # disable autocommit (isolation_level(1)) because DROP and
    # CREATE DATABASE can't be executed in a transaction
    set_isolation_level = getattr(cnx, "set_isolation_level", None)
    if set_isolation_level is not None:
        # set_isolation_level() is psycopg specific
        set_isolation_level(0)
    return cnx


def repo_cnx(config):
    """return a in-memory repository and a repoapi connection to it"""
    from cubicweb import repoapi
    from cubicweb.server.utils import manager_userpasswd

    db_driver = config.system_source_config.get("db-driver")
    if db_driver == "sqlite":
        if not os.path.exists(config.system_source_config.get("db-name")):
            print("Your database does not exist.")
            sys.exit(1)
        from sqlite3 import OperationalError
    else:
        from psycopg2 import OperationalError
    try:
        login = config.default_admin_config["login"]
        pwd = config.default_admin_config["password"]
    except KeyError:
        login, pwd = manager_userpasswd()
    while True:
        try:
            repo = repoapi.get_repository(config=config)
            cnx = repoapi.connect(repo, login, password=pwd)
            return repo, cnx
        except AuthenticationError:
            print("-> Error: wrong user/password.")
            # reset cubes else we'll have an assertion error on next retry
            config._cubes = None
        except (
            OperationalError
        ) as exc:  # sqlite or psycopg2 driver, depending on db_driver
            print(f"OperationalError: {exc}")
            sys.exit(1)
        login, pwd = manager_userpasswd()


# repository specific command handlers ########################################


class RepositoryCreateHandler(CommandHandler):
    cmdname = "create"
    cfgname = "repository"

    def bootstrap(self, cubes, automatic=False, inputlevel=0):
        """create an instance by copying files from the given cube and by asking
        information necessary to build required configuration files
        """
        config = self.config
        if not automatic:
            print(underline_title("Configuring the repository"))
            config.input_config("email", inputlevel)
            print("\n" + underline_title("Configuring the sources"))
        # hack to make Method('default_instance_id') usable in db option defs
        # (in native.py)
        sconfig = SourceConfiguration(config, options=SOURCE_TYPES["native"].options)
        if not automatic:
            sconfig.input_config(inputlevel=inputlevel)
            print()
        sourcescfg = {"system": sconfig}
        if automatic:
            # XXX modify a copy
            password = generate_password()
            print(f"-> set administrator account to admin / {password}")
            USER_OPTIONS[1][1]["default"] = password
            sconfig = Configuration(options=USER_OPTIONS)
        else:
            sconfig = Configuration(options=USER_OPTIONS)
            sconfig.input_config(inputlevel=inputlevel)
        sourcescfg["admin"] = sconfig
        config.write_sources_file(sourcescfg)
        # remember selected cubes for later initialization of the database
        config.write_bootstrap_cubes_file(cubes)

    def postcreate(self, automatic=False, inputlevel=0):
        if automatic:
            CWCTL.run(["db-create", "--automatic", self.config.appid])
        elif ASK.confirm("Run db-create to create the system database ?"):
            CWCTL.run(
                ["db-create", "--config-level", str(inputlevel), self.config.appid]
            )
        else:
            print(
                "-> nevermind, you can do it later with "
                '"cubicweb-ctl db-create %s".' % self.config.appid
            )


@contextmanager
def db_transaction(source, privilege):
    """Open a transaction to the instance database"""
    cnx = source_cnx(source, special_privs=privilege)
    cursor = cnx.cursor()
    try:
        yield cursor
    except Exception:
        cnx.rollback()
        cnx.close()
        raise
    else:
        cnx.commit()
        cnx.close()


@contextmanager
def db_sys_transaction(source, privilege):
    """Open a transaction to the system database"""
    cnx = _db_sys_cnx(source, privilege)
    cursor = cnx.cursor()
    try:
        yield cursor
    except Exception:
        cnx.rollback()
        cnx.close()
        raise
    else:
        cnx.commit()
        cnx.close()


class RepositoryDeleteHandler(CommandHandler):
    cmdname = "delete"
    cfgname = "repository"

    def _drop_namespace(self, source):
        db_namespace = source.get("db-namespace")
        with db_transaction(source, privilege="DROP SCHEMA") as cursor:
            helper = get_db_helper(source["db-driver"])
            helper.drop_schema(cursor, db_namespace)
            print(f"-> database schema {db_namespace} dropped")

    def _drop_database(self, source):
        if source["db-driver"] == "sqlite":
            print(f"deleting database file {source['db-name']}")
            os.unlink(source["db-name"])
            print(f"-> database {source['db-name']} dropped.")
        else:
            with db_sys_transaction(source, privilege="DROP DATABASE") as cursor:
                print(f"dropping database {source['db-name']}")
                cursor.execute(f"DROP DATABASE \"{source['db-name']}\"")
                print(f"-> database {source['db-name']} dropped.")

    def _drop_user(self, source):
        user = source["db-user"] or None
        if user is not None:
            with db_sys_transaction(source, privilege="DROP USER") as cursor:
                print(f"dropping user {user}")
                cursor.execute(f"DROP USER {user}")

    def _cleanup_steps(self, source):
        # 1/ delete namespace if used
        db_namespace = source.get("db-namespace")
        if db_namespace:
            yield (
                f'Delete database namespace "{db_namespace}"',
                self._drop_namespace,
                True,
            )
        # 2/ delete database
        yield (f"Delete database \"{source['db-name']}\"", self._drop_database, True)
        # 3/ delete user
        helper = get_db_helper(source["db-driver"])
        if source["db-user"] and helper.users_support:
            # XXX should check we are not connected as user
            yield (f"Delete user \"{source['db-user']}\"", self._drop_user, False)

    def cleanup(self):
        """remove instance's configuration and database"""
        source = self.config.system_source_config
        for msg, step, default in self._cleanup_steps(source):
            if ASK.confirm(msg, default_is_yes=default):
                try:
                    step(source)
                except Exception as exc:
                    print("ERROR", exc)
                    if ASK.confirm(
                        "An error occurred. Continue anyway?", default_is_yes=False
                    ):
                        continue
                    raise ExecutionError(str(exc))


# repository specific commands ################################################


def createdb(helper, source, dbcnx, cursor, **kwargs):
    if dbcnx.logged_user != source["db-user"]:
        helper.create_database(
            cursor,
            source["db-name"],
            source["db-user"],
            source["db-encoding"],
            **kwargs,
        )
    else:
        helper.create_database(
            cursor, source["db-name"], dbencoding=source["db-encoding"], **kwargs
        )


class CreateInstanceDBCommand(Command):
    """Create the system database of an instance (run after 'create').

    You will be prompted for a login / password to use to connect to
    the system database.  The given user should have almost all rights
    on the database (ie a super user on the DBMS allowed to create
    database, users, languages...).

    <instance>
      the identifier of the instance to initialize.
    """

    name = "db-create"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "automatic",
            {
                "short": "a",
                "action": "store_true",
                "default": False,
                "help": (
                    "automatic mode: never ask and use default answer to every "
                    "question. this may require that your login match a database super "
                    "user (allowed to create database & all). If the database already "
                    "exists and --drop is not provided, this command does nothing."
                ),
            },
        ),
        (
            "config-level",
            {
                "short": "l",
                "type": "int",
                "metavar": "<level>",
                "default": 0,
                "help": (
                    "configuration level (0..2): 0 will ask for essential "
                    "configuration parameters only while 2 will ask for all parameters"
                ),
            },
        ),
        (
            "create-db",
            {
                "short": "c",
                "type": "yn",
                "metavar": "<y or n>",
                "default": True,
                "help": "create the database (yes by default)",
            },
        ),
        (
            "drop",
            {
                "short": "d",
                "type": "yn",
                "metavar": "<y or n>",
                "default": None,
                "help": "Delete the database before creation if it exists (asked by default)",
            },
        ),
    )

    def run(self, args):
        """run the command with its specific arguments"""
        check_options_consistency(self.config)
        automatic = self.get("automatic")
        drop_db = self.get("drop")
        appid = args.pop()
        config = CubicWebConfiguration.config_for(appid)
        source = config.system_source_config
        dbname = source["db-name"]
        driver = source["db-driver"]
        helper = get_db_helper(driver)

        def should_drop_db():
            """Return True if the database should be dropped.

            The logic is following:
                - if drop_db is set then respect the user choice (either True or False)
                - if drop_db is not set then drop only in non automatic mode and
                  the user confirm the deletion
            """
            if drop_db is not None:
                return drop_db
            if automatic:
                return False
            drop_db_question = f"Database {dbname} already exists. Drop it?"
            return ASK.confirm(drop_db_question)

        if driver == "sqlite":
            if os.path.exists(dbname) and should_drop_db():
                os.unlink(dbname)
        elif self.config.create_db:
            print("\n" + underline_title("Creating the system database"))
            # connect on the dbms system base to create our base
            dbcnx = _db_sys_cnx(
                source, "CREATE/DROP DATABASE and / or USER", interactive=not automatic
            )
            cursor = dbcnx.cursor()
            try:
                if helper.users_support:
                    user = source["db-user"]
                    if not helper.user_exists(cursor, user) and (
                        automatic
                        or ASK.confirm(f"Create db user {user} ?", default_is_yes=False)
                    ):
                        helper.create_user(source["db-user"], source.get("db-password"))
                        print(f"-> user {user} created.")
                if dbname in helper.list_databases(cursor):
                    if should_drop_db():
                        cursor.execute(f'DROP DATABASE "{dbname}"')
                    else:
                        print(
                            "The database %s already exists, but automatically dropping it "
                            "is currently forbidden. You may want to run "
                            '"cubicweb-ctl db-create --drop=y %s" to continue or '
                            '"cubicweb-ctl db-create --help" to get help.'
                            % (dbname, config.appid)
                        )
                        raise Exception("Not allowed to drop existing database.")
                createdb(helper, source, dbcnx, cursor)
                dbcnx.commit()
                print(f"-> database {dbname} created.")
            except BaseException:
                dbcnx.rollback()
                raise
        cnx = source_cnx(
            source, special_privs="CREATE LANGUAGE/SCHEMA", interactive=not automatic
        )
        cursor = cnx.cursor()
        helper.init_fti_extensions(cursor)
        namespace = source.get("db-namespace")
        if namespace and (
            automatic
            or ASK.confirm(f"Create schema {namespace} in database {dbname} ?")
        ):
            helper.create_schema(cursor, namespace)
        cnx.commit()
        # postgres specific stuff
        if driver == "postgres":
            # install plpgsql language
            langs = ("plpgsql",)
            for extlang in langs:
                if automatic or ASK.confirm(f"Create language {extlang} ?"):
                    try:
                        helper.create_language(cursor, extlang)
                    except Exception as exc:
                        print("-> ERROR:", exc)
                        print(
                            "-> could not create language %s, "
                            "some stored procedures might be unusable" % extlang
                        )
                        cnx.rollback()
                    else:
                        cnx.commit()
        print(
            "-> database for instance %s created and necessary extensions installed."
            % appid
        )
        print()
        if automatic:
            CWCTL.run(["db-init", "--automatic", "--config-level", "0", config.appid])
        elif ASK.confirm("Run db-init to initialize the system database ?"):
            CWCTL.run(
                [
                    "db-init",
                    "--config-level",
                    str(self.config.config_level),
                    config.appid,
                ]
            )
        else:
            print(
                "-> nevermind, you can do it later with "
                '"cubicweb-ctl db-init %s".' % config.appid
            )


class InitInstanceCommand(Command):
    """Initialize the system database of an instance (run after 'db-create').

    Notice this will be done using user specified in the sources files, so this
    user should have the create tables grant permissions on the database.

    <instance>
      the identifier of the instance to initialize.
    """

    name = "db-init"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "automatic",
            {
                "short": "a",
                "action": "store_true",
                "default": False,
                "help": (
                    "automatic mode: never ask and use default answer to every "
                    "question."
                ),
            },
        ),
        (
            "config-level",
            {
                "short": "l",
                "type": "int",
                "default": 0,
                "help": (
                    "level threshold for questions asked when configuring "
                    "another source"
                ),
            },
        ),
        (
            "drop",
            {
                "short": "d",
                "action": "store_true",
                "default": False,
                "help": (
                    "insert drop statements to remove previously existant "
                    "tables, indexes... (no by default)"
                ),
            },
        ),
    )

    def run(self, args):
        check_options_consistency(self.config)
        print("\n" + underline_title("Initializing the system database"))
        from cubicweb.server import init_repository

        appid = args[0]
        config = CubicWebConfiguration.config_for(appid)
        try:
            system = config.system_source_config
            extra_args = system.get("db-extra-arguments")
            extra = extra_args and {"extra_args": extra_args} or {}
            get_connection(
                system["db-driver"],
                database=system["db-name"],
                host=system.get("db-host"),
                port=system.get("db-port"),
                user=system.get("db-user") or "",
                password=system.get("db-password") or "",
                schema=system.get("db-namespace"),
                **extra,
            )
        except Exception as ex:
            raise ConfigurationError(
                "You seem to have provided wrong connection information in "
                "the %s file. Resolve this first (error: %s)."
                % (config.sources_file(), str(ex).strip())
            )
        init_repository(config, drop=self.config.drop)
        if not self.config.automatic:
            while ASK.confirm("Enter another source ?", default_is_yes=False):
                CWCTL.run(
                    [
                        "source-add",
                        "--config-level",
                        str(self.config.config_level),
                        config.appid,
                    ]
                )


class AddSourceCommand(Command):
    """Add a data source to an instance.

    <instance>
      the identifier of the instance to initialize.
    """

    name = "source-add"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "config-level",
            {
                "short": "l",
                "type": "int",
                "default": 1,
                "help": "level threshold for questions asked when configuring another source",
            },
        ),
    )

    def run(self, args):
        appid = args[0]
        config = CubicWebConfiguration.config_for(appid)
        repo, cnx = repo_cnx(config)
        repo.hm.call_hooks("server_maintenance", repo=repo)
        try:
            with cnx:
                used = set(
                    n for n, in cnx.execute("Any SN WHERE S is CWSource, S name SN")
                )
                cubes = repo.get_cubes()
                while True:
                    type = input(f"source type ({', '.join(sorted(SOURCE_TYPES))}): ")
                    if type not in SOURCE_TYPES:
                        print("-> unknown source type, use one of the available types.")
                        continue
                    sourcemodule = SOURCE_TYPES[type].module
                    if not sourcemodule.startswith("cubicweb."):
                        # module names look like cubes.mycube.themodule
                        sourcecube = SOURCE_TYPES[type].module.split(".", 2)[1]
                        # if the source adapter is coming from an external component,
                        # ensure it's specified in used cubes
                        if sourcecube not in cubes:
                            print(
                                "-> this source type require the %s cube which is "
                                "not used by the instance."
                            )
                            continue
                    break
                while True:
                    parser = input(
                        f"parser type ({', '.join(sorted(repo.vreg['parsers']))}): "
                    )
                    if parser in repo.vreg["parsers"]:
                        break
                    print(
                        "-> unknown parser identifier, use one of the available types."
                    )
                while True:
                    sourceuri = input(
                        "source identifier (a unique name used to "
                        "tell sources apart): "
                    ).strip()
                    if not sourceuri:
                        print("-> mandatory.")
                    else:
                        sourceuri = str(sourceuri, sys.stdin.encoding)
                        if sourceuri in used:
                            print("-> uri already used, choose another one.")
                        else:
                            break
                url = input("source URL (leave empty for none): ").strip()
                url = str(url) if url else None
                # XXX configurable inputlevel
                sconfig = ask_source_config(
                    config, type, inputlevel=self.config.config_level
                )
                cfgstr = str(generate_source_config(sconfig), sys.stdin.encoding)
                cnx.create_entity(
                    "CWSource",
                    name=sourceuri,
                    type=str(type),
                    config=cfgstr,
                    parser=str(parser),
                    url=str(url),
                )
                cnx.commit()
        finally:
            repo.hm.call_hooks("server_shutdown")


class GrantUserOnInstanceCommand(Command):
    """Grant a database user on a repository system database.

    <instance>
      the identifier of the instance
    <user>
      the database's user requiring grant access
    """

    name = "db-grant-user"
    arguments = "<instance> <user>"
    min_args = max_args = 2
    options = (
        (
            "set-owner",
            {
                "short": "o",
                "type": "yn",
                "metavar": "<yes or no>",
                "default": False,
                "help": "Set the user as tables owner if yes (no by default).",
            },
        ),
    )

    def run(self, args):
        """run the command with its specific arguments"""
        from cubicweb.server.sqlutils import sqlexec, sqlgrants

        appid, user = args
        config = CubicWebConfiguration.config_for(appid)
        source = config.system_source_config
        set_owner = self.config.set_owner
        cnx = source_cnx(source, special_privs="GRANT")
        cursor = cnx.cursor()
        schema = config.load_schema()
        try:
            sqlexec(
                sqlgrants(schema, source["db-driver"], user, set_owner=set_owner),
                cursor,
            )
        except Exception as ex:
            cnx.rollback()
            import traceback

            traceback.print_exc()
            print("-> an error occurred:", ex)
        else:
            cnx.commit()
            print(f"-> rights granted to {appid} on instance {user}.")


class ResetAdminPasswordCommand(Command):
    """Reset the administrator password.

    <instance>
      the identifier of the instance
    """

    name = "reset-admin-pwd"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "password",
            {
                "short": "p",
                "type": "string",
                "metavar": "<new-password>",
                "default": None,
                "help": (
                    "Use this password instead of prompt for one.\n"
                    "/!\\ THIS IS AN INSECURE PRACTICE /!\\ \n"
                    "the password will appear in shell history"
                ),
            },
        ),
    )

    def run(self, args):
        """run the command with its specific arguments"""
        from cubicweb.server.utils import crypt_password, manager_userpasswd

        appid = args[0]
        config = CubicWebConfiguration.config_for(appid)
        sourcescfg = config.read_sources_file()
        try:
            adminlogin = sourcescfg["admin"]["login"]
        except KeyError:
            print("-> Error: could not get cubicweb administrator login.")
            sys.exit(1)
        cnx = source_cnx(sourcescfg["system"])
        driver = sourcescfg["system"]["db-driver"]
        dbhelper = get_db_helper(driver)
        cursor = cnx.cursor()
        # check admin exists
        cursor.execute(
            "SELECT * FROM cw_CWUser WHERE cw_login=%(l)s", {"l": adminlogin}
        )
        if not cursor.fetchall():
            print(
                "-> error: admin user %r specified in sources doesn't exist "
                "in the database" % adminlogin
            )
            print("   fix your sources file before running this command")
            cnx.close()
            sys.exit(1)
        if self.config.password is None:
            # ask for a new password
            msg = f"new password for {adminlogin}"
            _, pwd = manager_userpasswd(adminlogin, confirm=True, passwdmsg=msg)
        else:
            pwd = self.config.password
        try:
            cursor.execute(
                "UPDATE cw_CWUser SET cw_upassword=%(p)s WHERE cw_login=%(l)s",
                {"p": dbhelper.binary_value(crypt_password(pwd)), "l": adminlogin},
            )
            sconfig = Configuration(options=USER_OPTIONS)
            sconfig["login"] = adminlogin
            sconfig["password"] = pwd
            sourcescfg["admin"] = sconfig
            config.write_sources_file(sourcescfg)
        except Exception as ex:
            cnx.rollback()
            import traceback

            traceback.print_exc()
            print("-> an error occurred:", ex)
        else:
            cnx.commit()
            print("-> password reset, sources file regenerated.")
        cnx.close()


def _remote_dump(host, appid, output, sudo=False):
    # XXX generate unique/portable file name
    from datetime import date

    filename = f"{appid}-{date.today().strftime('%Y-%m-%d')}.tgz"
    dmpcmd = f"cubicweb-ctl db-dump -o /tmp/{filename} {appid}"
    if sudo:
        dmpcmd = f"sudo {dmpcmd}"
    dmpcmd = f'ssh -t {host} "{dmpcmd}"'
    print(dmpcmd)
    if os.system(dmpcmd):
        raise ExecutionError("Error while dumping the database")
    if output is None:
        output = filename
    cmd = f"scp {host}:/tmp/{filename} {output}"
    print(cmd)
    if os.system(cmd):
        raise ExecutionError(f"Error while retrieving the dump at /tmp/{filename}")
    rmcmd = f'ssh -t {host} "rm -f /tmp/{filename}"'
    print(rmcmd)
    if os.system(rmcmd) and not ASK.confirm(
        "An error occurred while deleting remote dump at /tmp/%s. "
        "Continue anyway?" % filename
    ):
        raise ExecutionError(f"Error while deleting remote dump at /tmp/{filename}")


def _local_dump(appid, output, format="native"):
    config = CubicWebConfiguration.config_for(appid)
    config.quick_start = True
    mih = config.migration_handler(verbosity=1)
    backupfile = mih.backup_database(output, askconfirm=False, format=format)
    mih.shutdown()
    return backupfile


def _get_s3_client(url, bucket, **kwargs):
    try:
        from minio import Minio
    except ImportError:
        raise ExecutionError(
            "Minio module is missing, s3 dump cannot be used. Install it with pip install .[s3]"
        )

    if not url:
        print("dump-s3-endpoint-url is not defined in all-in-one.conf")
        return

    # open s3 connection
    client = Minio(url, **kwargs)

    if not bucket:
        print("dump-s3-bucket-name is not defined in all-in-one.conf")
        return

    if not client.bucket_exists(bucket):
        print(f"dump-s3-bucket-name {bucket} do not exists")
        return

    return client


def _s3_dump(appid, output, clean, format="native"):
    config = CubicWebConfiguration.config_for(appid)
    s3_bucket_name = config.get("dump-s3-bucket-name")

    # open s3 connection
    client = _get_s3_client(
        config.get("dump-s3-endpoint-url"),
        s3_bucket_name,
        access_key=config.get("dump-s3-access-key"),
        secret_key=config.get("dump-s3-secret-key"),
        region=config.get("dump-s3-region"),
        secure=config.get("dump-s3-secure"),
        cert_check=config.get("dump-s3-cert-check"),
    )

    # create backup in tmp file
    backupfile = _local_dump(appid, None, format=format)
    backupname = os.path.basename(backupfile)
    backuppath = output if output else backupname

    # push to s3
    client.fput_object(s3_bucket_name, backuppath, backupfile)
    print(f"-> backup file {backupname} pushed to S3 bucket {s3_bucket_name}")

    # clean
    if clean or ASK.confirm(f"Delete local backup {backupname}?"):
        os.remove(backupfile)
        print(f"-> {backupname} has been deleted successfully")


def _local_restore(appid, backupfile, drop, format="native"):
    config = CubicWebConfiguration.config_for(appid)
    config.verbosity = 1  # else we won't be asked for confirmation on problems
    config.quick_start = True
    mih = config.migration_handler(connect=False, verbosity=1)
    mih.restore_database(backupfile, drop, askconfirm=False, format=format)
    repo = mih.repo
    # version of the database
    dbversions = repo.get_versions()
    mih.shutdown()
    if not dbversions:
        print(
            "bad or missing version information in the database, don't upgrade file system"
        )
        return
    # version of installed software
    eversion = dbversions["cubicweb"]
    status = instance_status(config, eversion, dbversions)
    # * database version > installed software
    if status == "needsoftupgrade":
        print(
            "** The database of %s is more recent than the installed software!"
            % config.appid
        )
        print(
            "** Upgrade your software, then migrate the database by running the command"
        )
        print(f"** 'cubicweb-ctl upgrade {config.appid}'")
        return
    # * database version < installed software, an upgrade will be necessary
    #   anyway, just rewrite vc.conf and warn user he has to upgrade
    elif status == "needapplupgrade":
        print(
            f"** The database of {config.appid} is older than the installed software."
        )
        print("** Migrate the database by running the command")
        print(f"** 'cubicweb-ctl upgrade {config.appid}'")
        return
    # * database version = installed software, database version = instance fs version
    #   ok!


def instance_status(config, cubicwebapplversion, vcconf):
    cubicwebversion = config.cubicweb_version()
    if cubicwebapplversion > cubicwebversion:
        return "needsoftupgrade"
    if cubicwebapplversion < cubicwebversion:
        return "needapplupgrade"
    for cube in config.cubes():
        try:
            softversion = config.cube_version(cube)
        except ConfigurationError:
            print(
                "-> Error: no cube version information for %s, "
                "please check that the cube is installed." % cube
            )
            continue
        try:
            applversion = vcconf[cube]
        except KeyError:
            print(
                "-> Error: no cube version information for %s in version configuration."
                % cube
            )
            continue
        if softversion == applversion:
            continue
        if softversion > applversion:
            return "needsoftupgrade"
        elif softversion < applversion:
            return "needapplupgrade"
    return None


class DBDumpCommand(Command):
    """Backup the system database of an instance.

    <instance>
      the identifier of the instance to backup
      format [[user@]host:]appname
    """

    name = "db-dump"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "output",
            {
                "short": "o",
                "type": "string",
                "metavar": "<file>",
                "default": None,
                "help": "Specify the backup file where the backup will be stored.",
            },
        ),
        (
            "s3",
            {
                "action": "store_true",
                "default": False,
                "help": "Push created local dump to S3 bucket.",
            },
        ),
        (
            "s3-clean",
            {
                "action": "store_true",
                "default": False,
                "help": "Delete local backup after created (and pushed to S3)",
            },
        ),
        (
            "sudo",
            {
                "short": "s",
                "action": "store_true",
                "default": False,
                "help": "Use sudo on the remote host.",
            },
        ),
        (
            "format",
            {
                "short": "f",
                "default": "native",
                "type": "choice",
                "choices": ("native", "portable"),
                "help": (
                    '"native" format uses db backend utilities to dump the database. '
                    '"portable" format uses a database independent format'
                ),
            },
        ),
    )

    def run(self, args):
        appid = args[0]
        if self.config.s3:
            _s3_dump(
                appid,
                self.config.output,
                self.config.s3_clean,
                format=self.config.format,
            )
        elif ":" in appid:
            host, appid = appid.split(":")
            _remote_dump(host, appid, self.config.output, self.config.sudo)
        else:
            _local_dump(appid, self.config.output, format=self.config.format)


class DBRestoreCommand(Command):
    """Restore the system database of an instance.

    <instance>
      the identifier of the instance to restore
    """

    name = "db-restore"
    arguments = "<instance> <backupfile>"
    min_args = 1
    max_args = 2

    options = (
        (
            "no-drop",
            {
                "short": "n",
                "action": "store_true",
                "default": False,
                "help": (
                    "for some reason the database doesn't exist and so "
                    "should not be dropped."
                ),
            },
        ),
        (
            "s3",
            {
                "action": "store_true",
                "default": False,
                "help": "Pull dump from S3 bucket.",
            },
        ),
        (
            "latest",
            {
                "action": "store_true",
                "default": False,
                "help": "Restore latest backup file of directory or s3 bucket.",
            },
        ),
        (
            "format",
            {
                "short": "f",
                "default": "native",
                "type": "choice",
                "choices": ("native", "portable"),
                "help": "the format used when dumping the database",
            },
        ),
    )

    def _get_latest_filename_from_s3_bucket(self, client, bucket_name):
        files = client.list_objects(bucket_name)
        latest_file = None
        for file in files:
            if not latest_file or file.last_modified > latest_file.last_modified:
                latest_file = file

        if not latest_file:
            raise ExecutionError(f"No file found in S3 bucket {bucket_name}")

        return latest_file.object_name

    def _get_latest_filename_from_path(self, directory):
        files = glob(f"{directory}/*")
        if not files:
            raise ExecutionError(f"No file found in {directory}")
        return max(files, key=os.path.getctime)

    def _get_dump_from_s3(self, appid, filepath, get_latest=False):
        # fget_object(bucket_name, object_name, file_path, request_headers=None, ssec=None, version_id=None, extra_query_params=None, tmp_file_path=None)
        config = CubicWebConfiguration.config_for(appid)
        s3_bucket_name = config.get("dump-s3-bucket-name")

        # open s3 connection
        client = _get_s3_client(
            config.get("dump-s3-endpoint-url"),
            s3_bucket_name,
            access_key=config.get("dump-s3-access-key"),
            secret_key=config.get("dump-s3-secret-key"),
            region=config.get("dump-s3-region"),
            secure=config.get("dump-s3-secure"),
            cert_check=config.get("dump-s3-cert-check"),
        )

        if get_latest:
            filepath = self._get_latest_filename_from_s3_bucket(client, s3_bucket_name)

        tmpfile = NamedTemporaryFile(suffix=f".{filepath.split('.', maxsplit=1)[-1]}")

        # download file to tmp file
        client.fget_object(s3_bucket_name, filepath, tmpfile.name)
        print(f"-> {filepath} has been downloaded as {tmpfile.name}")

        return tmpfile

    def run(self, args):
        appid = args[0]
        backupfile = None
        if len(args) == 2:
            backupfile = args[1]

        if self.config.format == "portable":
            # we need to ensure a DB exist before restoring from portable format
            if not self.config.no_drop:
                try:
                    CWCTL.run(["db-create", "--automatic", appid])
                except SystemExit as exc:
                    # continue if the command exited with status 0 (success)
                    if exc.code:
                        raise

        if self.config.s3:
            tmpfile = self._get_dump_from_s3(appid, backupfile, self.config.latest)
            backupfile = tmpfile.name

        # if --latest, backupfile must be a directory path.
        # get latest file in this directory
        if not self.config.s3 and self.config.latest:
            backupfile = self._get_latest_filename_from_path(backupfile)

        _local_restore(
            appid, backupfile, drop=not self.config.no_drop, format=self.config.format
        )

        # clean
        if self.config.s3:
            print(f"-> delete {tmpfile.name}")
            tmpfile.close()

        if self.config.format == "portable":
            try:
                CWCTL.run(["db-rebuild-fti", appid])
            except SystemExit as exc:
                if exc.code:
                    raise


class DBCopyCommand(Command):
    """Copy the system database of an instance (backup and restore).

    <src-instance>
      the identifier of the instance to backup
      format [[user@]host:]appname

    <dest-instance>
      the identifier of the instance to restore
    """

    name = "db-copy"
    arguments = "<src-instance> <dest-instance>"
    min_args = max_args = 2
    options = (
        (
            "no-drop",
            {
                "short": "n",
                "action": "store_true",
                "default": False,
                "help": (
                    "For some reason the database doesn't exist and so "
                    "should not be dropped."
                ),
            },
        ),
        (
            "keep-dump",
            {
                "short": "k",
                "action": "store_true",
                "default": False,
                "help": (
                    "Specify that the dump file should not be automatically removed."
                ),
            },
        ),
        (
            "sudo",
            {
                "short": "s",
                "action": "store_true",
                "default": False,
                "help": "Use sudo on the remote host.",
            },
        ),
        (
            "format",
            {
                "short": "f",
                "default": "native",
                "type": "choice",
                "choices": ("native", "portable"),
                "help": (
                    '"native" format uses db backend utilities to dump the database. '
                    '"portable" format uses a database independent format'
                ),
            },
        ),
    )

    def run(self, args):
        import tempfile

        srcappid, destappid = args
        fd, output = tempfile.mkstemp()
        os.close(fd)
        if ":" in srcappid:
            host, srcappid = srcappid.split(":")
            _remote_dump(host, srcappid, output, self.config.sudo)
        else:
            _local_dump(srcappid, output, format=self.config.format)
        _local_restore(destappid, output, not self.config.no_drop, self.config.format)
        if self.config.keep_dump:
            print("-> you can get the dump file at", output)
        else:
            os.remove(output)


class CheckRepositoryCommand(Command):
    """Check integrity of the system database of an instance.

    <instance>
      the identifier of the instance to check
    """

    name = "db-check"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "checks",
            {
                "short": "c",
                "type": "csv",
                "metavar": "<check list>",
                "default": sorted(checkintegrity._CHECKERS),
                "help": (
                    "Comma separated list of check to run. By default run all checks."
                ),
            },
        ),
        (
            "autofix",
            {
                "short": "a",
                "type": "yn",
                "metavar": "<yes or no>",
                "default": False,
                "help": 'Automatically correct integrity problems if this option \
is set to "y" or "yes", else only display them',
            },
        ),
        (
            "reindex",
            {
                "short": "r",
                "type": "yn",
                "metavar": "<yes or no>",
                "default": False,
                "help": 're-indexes the database for full text search if this \
option is set to "y" or "yes" (may be long for large database).',
            },
        ),
        (
            "force",
            {
                "short": "f",
                "action": "store_true",
                "default": False,
                "help": "don't check instance is up to date.",
            },
        ),
    )

    def run(self, args):
        appid = args[0]
        config = CubicWebConfiguration.config_for(appid)
        config.repairing = self.config.force
        repo, _cnx = repo_cnx(config)
        with repo.internal_cnx() as cnx:
            checkintegrity.check(
                repo, cnx, self.config.checks, self.config.reindex, self.config.autofix
            )


class DBIndexSanityCheckCommand(Command):
    """Check database indices of an instance.

    <instance>
      identifier of the instance to check
    """

    arguments = "<instance>"
    name = "db-check-index"
    min_args = 1

    def run(self, args):
        config = CubicWebConfiguration.config_for(args[0])
        repo, cnx = repo_cnx(config)
        with cnx:
            status = checkintegrity.check_indexes(cnx)
        sys.exit(status)


class DBCheckUnusedIndexCommand(Command):
    """Check all indexes in database and return unused ones since the last
    statistic reset (pg_stat_reset()).

    <instance>
      identifier of the instance to check
    """

    arguments = "<instance>"
    name = "db-check-unused-index"
    min_args = 1
    max_args = 1
    options = (
        (
            "min_usage",
            {
                "short": "m",
                "type": "int",
                "default": 0,
                "help": "The minimum usage count to consider an index as used.",
            },
        ),
    )

    def run(self, args):
        config = CubicWebConfiguration.config_for(args[0])
        repo, cnx = repo_cnx(config)
        with cnx:
            status = check_unused_index(cnx, self.config.min_usage)
        sys.exit(status)


class RebuildFTICommand(Command):
    """Rebuild the full-text index of the system database of an instance.

    <instance> [etype(s)]
      the identifier of the instance to rebuild

    If no etype is specified, cubicweb will reindex everything, otherwise
    only specified etypes will be considered.
    """

    name = "db-rebuild-fti"
    arguments = "<instance>"
    min_args = 1

    def run(self, args):
        from cubicweb.server.checkintegrity import reindex_entities

        appid = args.pop(0)
        etypes = args or None
        config = CubicWebConfiguration.config_for(appid)
        repo, cnx = repo_cnx(config)
        with cnx:
            reindex_entities(repo.schema, cnx, etypes=etypes)
            cnx.commit()


class RepositorySchedulerCommand(Command):
    """Start a repository tasks scheduler.

    Initialize a repository and start its tasks scheduler that would run
    registered "looping tasks".

    This is maintenance command that should be kept running along with a web
    instance of a CubicWeb WSGI application (e.g. embeded into a Pyramid
    application).

    <instance>
      the identifier of the instance
    """

    name = "scheduler"
    arguments = "<instance>"
    min_args = max_args = 1
    options = (
        (
            "loglevel",
            {
                "short": "l",
                "type": "choice",
                "metavar": "<log level>",
                "default": "info",
                "choices": ("debug", "info", "warning", "error"),
            },
        ),
    )

    def run(self, args):
        from cubicweb.cwctl import init_cmdline_log_threshold
        from cubicweb.server.repository import Repository

        config = CubicWebConfiguration.config_for(args[0])
        # Log to stdout, since the this command runs in the foreground.
        config.global_set_option("log-file", None)
        init_cmdline_log_threshold(config, self["loglevel"])
        repo = Repository(config, sched.scheduler())
        repo.bootstrap()
        try:
            repo.run_scheduler()
        finally:
            repo.shutdown()


class SynchronizeSourceCommand(Command):
    """Force sources synchronization.

    <instance>
      the identifier of the instance
    <source>
      names of the sources to synchronize, if empty all sources will be synced.
    """

    name = "source-sync"
    arguments = "<instance> [<source> <source> ...]"
    min_args = 1
    options = (
        (
            "loglevel",
            {
                "short": "l",
                "type": "choice",
                "metavar": "<log level>",
                "default": "info",
                "choices": ("debug", "info", "warning", "error"),
            },
        ),
        (
            "force",
            {
                "short": "f",
                "action": "store_true",
                "default": False,
                "help": (
                    "force source synchronization (ignore synchronization interval)"
                ),
            },
        ),
    )

    def run(self, args):
        from cubicweb import repoapi
        from cubicweb.cwctl import init_cmdline_log_threshold

        config = CubicWebConfiguration.config_for(args[0])
        config.global_set_option("log-file", None)
        config.log_format = "%(levelname)s %(name)s: %(message)s"
        init_cmdline_log_threshold(config, self["loglevel"])
        repo = repoapi.get_repository(config=config)
        repo.hm.call_hooks("server_maintenance", repo=repo)
        errors = False
        with repo.internal_cnx() as cnx:
            sources = []
            if len(args) >= 2:
                for name in args[1:]:
                    try:
                        source = repo.source_by_uri(name)
                    except ValueError:
                        cnx.error(f"no source named {name!r}")
                        errors = True
                    else:
                        sources.append(source)
            else:
                for uri, source in repo.sources_by_uri.items():
                    if (
                        uri != "system"
                        and repo.config.source_enabled(source)
                        and source.config["synchronize"]
                    ):
                        sources.append(source)

            for source in sources:
                try:
                    stats = source.pull_data(
                        cnx, force=self["force"], raise_on_error=True
                    )
                except Exception:
                    cnx.exception("while trying to update source %s", source)
                    errors = True
                else:
                    for key, val in stats.items():
                        if val:
                            print(key, ":", val)

        if errors:
            raise ExecutionError("All sources where not synced")


def permissionshandler(relation, perms):
    from yams.buildobjs import DEFAULT_ATTRPERMS
    from cubicweb.schema import (
        PUB_SYSTEM_ENTITY_PERMS,
        PUB_SYSTEM_REL_PERMS,
        PUB_SYSTEM_ATTR_PERMS,
        RO_REL_PERMS,
        RO_ATTR_PERMS,
    )

    defaultrelperms = (
        DEFAULT_ATTRPERMS,
        PUB_SYSTEM_REL_PERMS,
        PUB_SYSTEM_ATTR_PERMS,
        RO_REL_PERMS,
        RO_ATTR_PERMS,
    )
    defaulteperms = (PUB_SYSTEM_ENTITY_PERMS,)
    # canonicalize vs str/unicode
    for p in ("read", "add", "update", "delete"):
        rule = perms.get(p)
        if rule:
            perms[p] = tuple(rule)
    return perms, perms in defaultrelperms or perms in defaulteperms


class SchemaDiffCommand(Command):
    """Generate a diff between schema and fsschema description.

    <instance>
      the identifier of the instance
    <diff-tool>
      the name of the diff tool to compare the two generated files.
    """

    name = "schema-diff"
    arguments = "<instance> <diff-tool>"
    min_args = max_args = 2

    def run(self, args):
        from yams.diff import schema_diff
        from cubicweb import repoapi

        appid = args.pop(0)
        diff_tool = args.pop(0)
        config = CubicWebConfiguration.config_for(appid)
        config.repairing = True
        repo = repoapi.get_repository(config=config)
        fsschema = config.load_schema(expand_cubes=True)
        schema_diff(
            fsschema, repo.schema, permissionshandler, diff_tool, ignore=("eid",)
        )


for cmdclass in (
    CreateInstanceDBCommand,
    InitInstanceCommand,
    GrantUserOnInstanceCommand,
    ResetAdminPasswordCommand,
    DBDumpCommand,
    DBRestoreCommand,
    DBCopyCommand,
    DBIndexSanityCheckCommand,
    DBCheckUnusedIndexCommand,
    AddSourceCommand,
    CheckRepositoryCommand,
    RebuildFTICommand,
    SynchronizeSourceCommand,
    SchemaDiffCommand,
    RepositorySchedulerCommand,
):
    CWCTL.register(cmdclass)

# extend configure command to set options in sources config file ###############

db_options = (
    (
        "db",
        {
            "short": "d",
            "type": "named",
            "metavar": "[section1.]key1:value1,[section2.]key2:value2",
            "default": None,
            "help": (
                """set <key> in <section> to <value> in "source" configuration file. If
<section> is not specified, it defaults to "system".

Beware that changing admin.login or admin.password using this command
will NOT update the database with new admin credentials.  Use the
reset-admin-pwd command instead.
"""
            ),
        },
    ),
)

ConfigureInstanceCommand.options = merge_options(
    ConfigureInstanceCommand.options + db_options
)

configure_instance = ConfigureInstanceCommand.configure_instance


def configure_instance2(self, appid):
    configure_instance(self, appid)
    if self.config.db is not None:
        appcfg = CubicWebConfiguration.config_for(appid)
        srccfg = appcfg.read_sources_file()
        for key, value in self.config.db.items():
            if "." in key:
                section, key = key.split(".", 1)
            else:
                section = "system"
            try:
                srccfg[section][key] = value
            except KeyError:
                raise ConfigurationError(
                    'unknown configuration key "%s" in section "%s" for source'
                    % (key, section)
                )
        admcfg = Configuration(options=USER_OPTIONS)
        admcfg["login"] = srccfg["admin"]["login"]
        admcfg["password"] = srccfg["admin"]["password"]
        srccfg["admin"] = admcfg
        appcfg.write_sources_file(srccfg)


ConfigureInstanceCommand.configure_instance = configure_instance2



from yams.buildobjs import ComputedRelation, EntityType, RelationDefinition

from cubicweb.schema import RRQLExpression


class Subject(EntityType):
    pass


class Object(EntityType):
    pass


class relation(RelationDefinition):
    subject = "Subject"
    object = "Object"


class computed(ComputedRelation):
    rule = "S relation O"
    __permissions__ = {"read": (RRQLExpression("S is ET"),)}



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
"""cubicweb-geonames schema

See geonames readme.txt for more details.
"""

from yams.buildobjs import EntityType, SubjectRelation, String, Int, BigInt, Float


class Location(EntityType):
    """
    Entity type for location of Geonames.
    See cities1000.zip, cities5000.zip, cities15000.zip and allCountries.txt
    """

    name = String(maxsize=1024, indexed=True, fulltextindexed=True)
    geonameid = Int(indexed=True)
    asciiname = String(maxsize=200, fulltextindexed=True)
    alternatenames = String(fulltextindexed=True)
    latitude = Float(indexed=True)
    longitude = Float(indexed=True)
    feature_class = String(maxsize=1, indexed=True)
    alternate_country_code = String(maxsize=60)
    admin_code_3 = String(maxsize=20)
    admin_code_4 = String(maxsize=20)
    population = BigInt(indexed=True)
    elevation = Int(indexed=True)
    gtopo30 = Int(indexed=True)
    timezone = SubjectRelation("TimeZone", cardinality="?*", inlined=True)


class TimeZone(EntityType):
    """
    Entity type for timezone of geonames.
    See timeZones.txt
    """

    code = String(maxsize=1024, indexed=True, required=True)
    gmt = Float()
    dst = Float()
    raw_offset = Float()



# copyright 2010-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some basic entity adapter implementations, for interfaces used in the
framework itself.
"""
from hashlib import sha1
from itertools import chain

from logilab.common.decorators import cached, cachedproperty
from logilab.mtconverter import TransformError
from rdflib import URIRef, Literal, namespace as rdflib_namespace

from yams import ValidationError
from cubicweb import (
    Unauthorized,
    ViolatedConstraint,
    UniqueTogetherError,
)
from cubicweb import _
from cubicweb import rdf
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import is_instance, relation_possible, match_exception
from cubicweb.schema import constraint_name_for, VIRTUAL_RTYPES

_marker = object()


class EntityRDFAdapter(EntityAdapter):
    """EntityRDFAdapter is to be specialized for each entity that wants to
    be converted to RDF using the mechanism from cubicweb.rdf
    """

    __regid__ = "rdf"

    SKIP_RTYPES = VIRTUAL_RTYPES | set(["cwuri", "is", "is_instance_of"])

    def __init__(self, _cw, **kwargs):
        super().__init__(_cw, **kwargs)
        self.entity.complete()
        self.used_namespaces = {}

    def _use_namespace(self, prefix, base_url=None):
        if prefix not in self.used_namespaces:
            if base_url is not None:
                if prefix in rdf.NAMESPACES:
                    raise KeyError(
                        "prefix redefinition not allowed: "
                        f'"{prefix}" already exists as "{rdf.NAMESPACES[prefix]}"'
                    )
                ns = rdflib_namespace.Namespace(base_url)
            else:
                ns = rdf.NAMESPACES[prefix]
            self.used_namespaces[prefix] = ns
        elif base_url is not None:
            if self.used_namespaces[prefix] != rdflib_namespace.Namespace(base_url):
                raise ValueError(
                    "prefix redefinition not allowed: "
                    f'"{prefix}" already exists as "{self.used_namespaces[prefix]}"'
                )
        return self.used_namespaces[prefix]

    @cachedproperty
    def uri(self):
        return URIRef(self.entity.absolute_url())

    def triples(self):
        """return sequence of 3-tuple of rdflib identifiers"""
        yield from self.cw_triples()
        yield from self.dc_triples()

    def cw_triples(self):
        RDF = self._use_namespace("rdf")
        OWL = self._use_namespace("owl")
        CW = self._use_namespace("cubicweb")

        yield (self.uri, RDF.type, CW[self.entity.e_schema.type])
        yield (self.uri, OWL.sameAs, URIRef(self.entity.cwuri))

        for rschema, targettypes, role in self.entity.e_schema.relation_definitions(
            "relation"
        ):
            rtype = rschema.type
            if rtype in self.SKIP_RTYPES or rtype.endswith("_permission"):
                continue
            for targetype in targettypes:
                # if rschema is an attribute
                if targetype.final:
                    try:
                        value = self.entity.cw_attr_cache[rtype]
                    except KeyError:
                        continue
                    if value is not None:
                        yield (self.uri, CW[rtype], Literal(value))
                # else if rschema is a relation
                else:
                    for related in self.entity.related(
                        rtype, role, entities=True, safe=True
                    ):
                        if role == "subject":
                            yield (self.uri, CW[rtype], URIRef(related.cwuri))
                        else:
                            yield (URIRef(related.cwuri), CW[rtype], self.uri)

    def dc_triples(self):
        dc_entity = self.entity.cw_adapt_to("IDublinCore")
        DC = self._use_namespace("dc")
        yield (self.uri, DC.title, Literal(dc_entity.long_title()))  # or title() ?
        desc = dc_entity.description()
        if desc:
            yield (self.uri, DC.description, Literal(desc))
        creator = dc_entity.creator()  # use URI instead of Literal ?
        if creator:
            yield (self.uri, DC.creator, Literal(creator))
        yield (self.uri, DC.date, Literal(dc_entity.date()))
        yield (self.uri, DC.type, Literal(dc_entity.type()))
        yield (self.uri, DC.language, Literal(dc_entity.language()))


class CWUserRDFAdapter(EntityRDFAdapter):
    __select__ = is_instance("CWUser")

    def triples(self):
        yield from super().triples()
        yield from self.foaf_triples()

    def foaf_triples(self):
        RDF = self._use_namespace("rdf")
        FOAF = self._use_namespace("foaf")
        yield (self.uri, RDF.type, FOAF.Person)
        if self.entity.surname:
            yield (self.uri, FOAF.familyName, Literal(self.entity.surname))
        if self.entity.firstname:
            yield (self.uri, FOAF.givenName, Literal(self.entity.firstname))
        emailaddr = self.entity.cw_adapt_to("IEmailable").get_email()
        if emailaddr:
            email_digest = sha1(emailaddr.encode("utf-8")).hexdigest()
            yield (self.uri, FOAF.mbox_sha1sum, Literal(email_digest))


class IDublinCoreAdapter(EntityAdapter):
    __regid__ = "IDublinCore"
    __select__ = is_instance("Any")

    def title(self):
        """Return a suitable *unicode* title for entity"""
        entity = self.entity
        for rschema, attrschema in entity.e_schema.attribute_definitions():
            if rschema.meta:
                continue
            value = entity.cw_attr_value(rschema.type)
            if value is not None:
                # make the value printable (dates, floats, bytes, etc.)
                return entity.printable_value(
                    rschema.type, value, attrschema.type, format="text/plain"
                )
        return f"{self.type()} #{entity.eid}"

    def long_title(self):
        """Return a more detailled title for entity"""
        # go through entity.dc_title for bw compat purpose: if entity define dc_title but not
        # dc_long_title, we still want it to be considered.
        return self.entity.dc_title()

    def description(self, format="text/plain"):
        """Return a suitable description for entity"""
        if "description" in self.entity.e_schema.subject_relations:
            return self.entity.printable_value("description", format=format)
        return ""

    def authors(self):  # XXX is this part of DC ?
        """Return a suitable description for the author(s) of the entity"""
        try:
            return ", ".join(u.name() for u in self.entity.owned_by)
        except Unauthorized:
            return ""

    def creator(self):
        """Return a suitable description for the creator of the entity"""
        if self.entity.creator:
            return self.entity.creator.name()
        return ""

    def date(self, date_format=None):  # XXX default to ISO 8601 ?
        """Return latest modification date of entity"""
        return self._cw.format_date(
            self.entity.modification_date, date_format=date_format
        )

    def type(self, form=""):
        """Return the display name for the type of entity (translated)"""
        return self.entity.e_schema.display_name(self._cw, form)

    def language(self):
        """Return language used by this entity (translated)"""
        eschema = self.entity.e_schema
        # check if entities has internationalizable attributes
        # XXX one is enough or check if all String attributes are internationalizable?
        for rschema, attrschema in eschema.attribute_definitions():
            if getattr(
                rschema.relation_definition(eschema, attrschema),
                "internationalizable",
                False,
            ):
                return self._cw._(self._cw.user.property_value("ui.language"))
        return self._cw._(self._cw.vreg.property_value("ui.language"))


class IEmailableAdapter(EntityAdapter):
    __regid__ = "IEmailable"
    __select__ = relation_possible("primary_email") | relation_possible("use_email")

    def get_email(self):
        if getattr(self.entity, "primary_email", None):
            return self.entity.primary_email[0].address
        if getattr(self.entity, "use_email", None):
            return self.entity.use_email[0].address
        return None

    def allowed_massmail_keys(self):
        """returns a set of allowed email substitution keys

        The default is to return the entity's attribute list but you might
        override this method to allow extra keys.  For instance, a Person
        class might want to return a `companyname` key.
        """
        return set(
            rschema.type
            for rschema, attrtype in self.entity.e_schema.attribute_definitions()
            if attrtype.type not in ("Password", "Bytes")
        )

    def as_email_context(self):
        """returns the dictionary as used by the sendmail controller to
        build email bodies.

        NOTE: the dictionary keys should match the list returned by the
        `allowed_massmail_keys` method.
        """
        return dict(
            (attr, getattr(self.entity, attr)) for attr in self.allowed_massmail_keys()
        )


class INotifiableAdapter(EntityAdapter):
    __regid__ = "INotifiable"
    __select__ = is_instance("Any")

    def notification_references(self, view):
        """used to control References field of email send on notification
        for this entity. `view` is the notification view.

        Should return a list of eids which can be used to generate message
        identifiers of previously sent email(s)
        """
        itree = self.entity.cw_adapt_to("ITree")
        if itree is not None:
            return itree.path()[:-1]
        if view.msgid_timestamp:
            return (self.entity.eid,)
        return ()


class IFTIndexableAdapter(EntityAdapter):
    """standard adapter to handle fulltext indexing

    .. automethod:: cubicweb.entities.adapters.IFTIndexableAdapter.fti_containers
    .. automethod:: cubicweb.entities.adapters.IFTIndexableAdapter.get_words
    """

    __regid__ = "IFTIndexable"
    __select__ = is_instance("Any")

    def fti_containers(self, _done=None):
        """return the list of entities to index when handling ``self.entity``

        The actual list of entities depends on ``fulltext_container`` usage
        in the datamodel definition
        """
        if _done is None:
            _done = set()
        entity = self.entity
        _done.add(entity.eid)
        containers = tuple(entity.e_schema.fulltext_containers())
        if containers:
            for rschema, role in containers:
                if role == "object":
                    targets = getattr(entity, rschema.type)
                else:
                    targets = getattr(entity, f"reverse_{rschema}")
                for target in targets:
                    if target.eid in _done:
                        continue
                    for container in target.cw_adapt_to("IFTIndexable").fti_containers(
                        _done
                    ):
                        yield container
        else:
            yield entity

    # weight in ABCD
    entity_weight = 1.0
    attr_weight = {}

    def get_words(self):
        """used by the full text indexer to get words to index

        this method should only be used on the repository side since it depends
        on the logilab.database package

        :rtype: list
        :return: the list of indexable word of this entity
        """
        from logilab.database.fti import tokenize

        # take care to cases where we're modyfying the schema
        entity = self.entity
        pending = self._cw.transaction_data.setdefault("pendingrdefs", set())
        words = {}
        for rschema in entity.e_schema.indexable_attributes():
            if (entity.e_schema, rschema) in pending:
                continue
            weight = self.attr_weight.get(rschema, "C")
            try:
                value = entity.printable_value(rschema, format="text/plain")
            except TransformError:
                continue
            except Exception:
                self.exception(
                    "can't add value of %s to text index for entity %s",
                    rschema,
                    entity.eid,
                )
                continue
            if value:
                words.setdefault(weight, []).extend(tokenize(value))
        for rschema, role in entity.e_schema.fulltext_relations():
            if role == "subject":
                for entity_ in getattr(entity, rschema.type):
                    merge_weight_dict(
                        words, entity_.cw_adapt_to("IFTIndexable").get_words()
                    )
            else:  # if role == 'object':
                for entity_ in getattr(entity, f"reverse_{rschema.type}"):
                    merge_weight_dict(
                        words, entity_.cw_adapt_to("IFTIndexable").get_words()
                    )
        return words


def merge_weight_dict(maindict, newdict):
    for weight, words in newdict.items():
        maindict.setdefault(weight, []).extend(words)


class IDownloadableAdapter(EntityAdapter):
    """interface for downloadable entities"""

    __regid__ = "IDownloadable"
    __abstract__ = True

    def download_url(self, **kwargs):  # XXX not really part of this interface
        """return a URL to download entity's content

        It should be a unicode object containing url-encoded ASCII.
        """
        raise NotImplementedError

    def download_content_type(self):
        """return MIME type (unicode) of the downloadable content"""
        raise NotImplementedError

    def download_encoding(self):
        """return encoding (unicode) of the downloadable content"""
        raise NotImplementedError

    def download_file_name(self):
        """return file name (unicode) of the downloadable content"""
        raise NotImplementedError

    def download_data(self):
        """return actual data (bytes) of the downloadable content"""
        raise NotImplementedError


# XXX should propose to use two different relations for children/parent
class ITreeAdapter(EntityAdapter):
    """This adapter provides a tree interface.

    It has to be overriden to be configured using the tree_relation,
    child_role and parent_role class attributes to benefit from this default
    implementation.

    This class provides the following methods:

    .. automethod: iterparents
    .. automethod: iterchildren
    .. automethod: prefixiter

    .. automethod: is_leaf
    .. automethod: is_root

    .. automethod: root
    .. automethod: parent
    .. automethod: children
    .. automethod: different_type_children
    .. automethod: same_type_children
    .. automethod: children_rql
    .. automethod: path
    """

    __regid__ = "ITree"
    __abstract__ = True

    child_role = "subject"
    parent_role = "object"

    def children_rql(self):
        """Returns RQL to get the children of the entity."""
        return self.entity.cw_related_rql(self.tree_relation, self.parent_role)

    def different_type_children(self, entities=True):
        """Return children entities of different type as this entity.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        res = self.entity.related(
            self.tree_relation, self.parent_role, entities=entities
        )
        eschema = self.entity.e_schema
        if entities:
            return [e for e in res if e.e_schema != eschema]
        return res.filtered_rset(lambda x: x.e_schema != eschema, self.entity.cw_col)

    def same_type_children(self, entities=True):
        """Return children entities of the same type as this entity.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        res = self.entity.related(
            self.tree_relation, self.parent_role, entities=entities
        )
        eschema = self.entity.e_schema
        if entities:
            return [e for e in res if e.e_schema == eschema]
        return res.filtered_rset(lambda x: x.e_schema is eschema, self.entity.cw_col)

    def is_leaf(self):
        """Returns True if the entity does not have any children."""
        return len(self.children()) == 0

    def is_root(self):
        """Returns true if the entity is root of the tree (e.g. has no parent)."""
        return self.parent() is None

    def root(self):
        """Return the root entity of the tree."""
        return self._cw.entity_from_eid(self.path()[0])

    def parent(self):
        """Returns the parent entity if any, else None (e.g. if we are on the
        root).
        """
        try:
            return self.entity.related(
                self.tree_relation, self.child_role, entities=True
            )[0]
        except (KeyError, IndexError):
            return None

    def children(self, entities=True, sametype=False):
        """Return children entities.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        if sametype:
            return self.same_type_children(entities)
        else:
            return self.entity.related(
                self.tree_relation, self.parent_role, entities=entities
            )

    def iterparents(self, strict=True):
        """Return an iterator on the parents of the entity."""

        def _uptoroot(self):
            curr = self
            while True:
                curr = curr.parent()
                if curr is None:
                    break
                yield curr
                curr = curr.cw_adapt_to("ITree")

        if not strict:
            return chain([self.entity], _uptoroot(self))
        return _uptoroot(self)

    def iterchildren(self, _done=None):
        """Return an iterator over the item's children."""
        if _done is None:
            _done = set()
        for child in self.children():
            if child.eid in _done:
                self.error("loop in %s tree: %s", child.cw_etype.lower(), child)
                continue
            yield child
            _done.add(child.eid)

    def prefixiter(self, _done=None):
        """Return an iterator over the item's descendants in a prefixed order."""
        if _done is None:
            _done = set()
        if self.entity.eid in _done:
            return
        _done.add(self.entity.eid)
        yield self.entity
        for child in self.same_type_children():
            for entity in child.cw_adapt_to("ITree").prefixiter(_done):
                yield entity

    @cached
    def path(self):
        """Returns the list of eids from the root object to this object."""
        path = []
        adapter = self
        entity = adapter.entity
        while entity is not None:
            if entity.eid in path:
                self.error("loop in %s tree: %s", entity.cw_etype.lower(), entity)
                break
            path.append(entity.eid)
            try:
                # check we are not jumping to another tree
                if (
                    adapter.tree_relation != self.tree_relation
                    or adapter.child_role != self.child_role
                ):
                    break
                entity = adapter.parent()
                adapter = entity.cw_adapt_to("ITree")
            except AttributeError:
                break
        path.reverse()
        return path


class ISerializableAdapter(EntityAdapter):
    """Adapter to serialize an entity to a bare python structure that may be
    directly serialized to e.g. JSON.
    """

    __regid__ = "ISerializable"
    __select__ = is_instance("Any")

    def serialize(self):
        entity = self.entity
        entity.complete()
        data = {
            "cw_etype": entity.cw_etype,
            "eid": entity.eid,
        }
        for rschema, __ in entity.e_schema.attribute_definitions():
            attr = rschema.type
            try:
                value = entity.cw_attr_cache[attr]
            except KeyError:
                # Bytes
                continue
            data[attr] = value
        return data


# error handling adapters ######################################################


class IUserFriendlyError(EntityAdapter):
    __regid__ = "IUserFriendlyError"
    __abstract__ = True

    def __init__(self, *args, **kwargs):
        self.exc = kwargs.pop("exc")
        super(IUserFriendlyError, self).__init__(*args, **kwargs)


class IUserFriendlyUniqueTogether(IUserFriendlyError):
    __select__ = match_exception(UniqueTogetherError)

    def raise_user_exception(self):
        rtypes = self.exc.rtypes
        errors = {}
        msgargs = {}
        i18nvalues = []
        for rtype in rtypes:
            errors[rtype] = _("%(KEY-rtype)s is part of violated unicity constraint")
            msgargs[rtype + "-rtype"] = rtype
            i18nvalues.append(rtype + "-rtype")
        errors[""] = _("some relations violate a unicity constraint")
        raise ValidationError(
            self.entity.eid, errors, msgargs=msgargs, i18nvalues=i18nvalues
        )


class IUserFriendlyCheckConstraint(IUserFriendlyError):
    __select__ = match_exception(ViolatedConstraint)

    def raise_user_exception(self):
        cstrname = self.exc.cstrname
        eschema = self.entity.e_schema
        for rschema, attrschema in eschema.attribute_definitions():
            rdef = rschema.relation_definition(eschema, attrschema)
            for constraint in rdef.constraints:
                if cstrname == constraint_name_for(constraint, rdef):
                    break
            else:
                continue
            break
        else:
            assert 0
        key = rschema.type + "-subject"
        # use .get since a constraint may be associated to an attribute that isn't edited (e.g.
        # constraint between two attributes). This should be the purpose of an api rework at some
        # point, we currently rely on the fact that such constraint will provide a dedicated user
        # message not relying on the `value` argument
        value = self.entity.cw_edited.get(rschema.type)
        msg, args = constraint.failed_message(key, value, self.entity)
        raise ValidationError(self.entity.eid, {key: msg}, args)


class EntityRQLInterfaceAdapter(EntityAdapter):
    __regid__ = "IRQLInterface"

    def absolute_url(self):
        return self.entity.absolute_url()

    def rest_path(self):
        return self.entity.rest_path()

    def cw_attr_metadata(self, attr, metadata):
        return self.entity.cw_attr_metadata(attr, metadata)

    def printable_value(
        self, attr, value=_marker, attrtype=None, format="text/html", displaytime=True
    ):
        if value is _marker:
            return self.entity.printable_value(
                attr, attrtype=attrtype, format=format, displaytime=displaytime
            )
        return self.entity.printable_value(attr, value, attrtype, format, displaytime)

    def dc_title(self):
        return self.entity.dc_title()

    def dc_long_title(self):
        return self.entity.dc_long_title()

    def dc_description(self):
        return self.entity.dc_description()

    def dc_authors(self):
        return self.entity.dc_authors()

    def dc_creator(self):
        return self.entity.dc_creator()

    def dc_date(self):
        return self.entity.dc_date()

    def dc_type(self):
        return self.entity.dc_type()

    def dc_language(self):
        return self.entity.dc_language()

    def creator(self):
        return self.entity.creator



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import (
    EntityType,
    String,
    RichString,
    Bytes,
    Int,
    ComputedRelation,
    SubjectRelation,
    RelationDefinition,
)

from cubicweb import _
from cubicweb.schema import (
    WorkflowableEntityType,
    RQLConstraint,
    RQLVocabularyConstraint,
)


class buddies(ComputedRelation):
    rule = "S in_group G, O in_group G"


class Personne(EntityType):
    nom = String(required=True)
    prenom = String()
    type = String()
    travaille = SubjectRelation("Societe")
    evaluee = SubjectRelation(("Note", "Personne"))
    connait = SubjectRelation(
        "Personne",
        symmetric=True,
        constraints=[
            RQLConstraint("NOT S identity O"),
            # conflicting constraints, see cw_unrelated_rql tests in
            # unittest_entity.py
            RQLVocabularyConstraint('NOT (S connait P, P nom "toto")'),
            RQLVocabularyConstraint('S travaille P, P nom "tutu"'),
        ],
    )
    actionnaire = SubjectRelation(
        "Societe",
        cardinality="??",
        constraints=[RQLConstraint("NOT EXISTS(O contrat_exclusif S)")],
    )
    dirige = SubjectRelation(
        "Societe", cardinality="??", constraints=[RQLConstraint("S actionnaire O")]
    )
    associe = SubjectRelation(
        "Personne",
        cardinality="?*",
        constraints=[RQLConstraint("S actionnaire SOC, O actionnaire SOC")],
    )
    promo = String(vocabulary=("bon", "pasbon"))


class Ami(EntityType):
    """A Person, for which surname is not required"""

    prenom = String()
    nom = String()


class Ville(EntityType):
    insee = Int(required=True)


class Societe(EntityType):
    nom = String()
    evaluee = SubjectRelation("Note")
    fournit = SubjectRelation(("Service", "Produit"), cardinality="1*")
    contrat_exclusif = SubjectRelation("Personne", cardinality="??")


class Service(EntityType):
    fabrique_par = SubjectRelation("Personne", cardinality="1*")


class Produit(EntityType):
    fabrique_par = SubjectRelation("Usine", cardinality="1*", inlined=True)


class Usine(EntityType):
    lieu = String(required=True)


class Note(EntityType):
    type = String()
    ecrit_par = SubjectRelation("Personne")


class SubNote(Note):
    __specializes_schema__ = True
    description = String()


class tags(RelationDefinition):
    subject = "Tag"
    object = ("Ami", "Personne", "Note")


class evaluee(RelationDefinition):
    subject = "CWUser"
    object = "Note"


class StateFull(WorkflowableEntityType):
    name = String()


class Reference(EntityType):
    nom = String(unique=True)
    ean = String(unique=True, required=True)


class FakeFile(EntityType):
    title = String(fulltextindexed=True, maxsize=256)
    data = Bytes(required=True, fulltextindexed=True, description=_("file to upload"))
    data_format = String(
        required=True,
        maxsize=128,
        description=_(
            "MIME type of the file. Should be dynamically set at upload time."
        ),
    )
    data_encoding = String(
        maxsize=32,
        description=_(
            "encoding of the file when it applies (e.g. text). "
            "Should be dynamically set at upload time."
        ),
    )
    data_name = String(
        required=True,
        fulltextindexed=True,
        description=_("name of the file. Should be dynamically set at upload time."),
    )
    description = RichString(
        fulltextindexed=True, internationalizable=True, default_format="text/rest"
    )



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Various tools.

.. warning::

    This module should be considered as internal implementation details. Use
    with caution, as the API may change without notice.
"""

from repoze.lru import lru_cache


def clone_user(repo, user):
    """Clone a CWUser instance.

    .. warning::

        The returned clone is detached from any cnx.
        Before using it in any way, it should be attached to a cnx that has not
        this user already loaded.
    """
    CWUser = repo.vreg["etypes"].etype_class("CWUser")
    clone = CWUser(None, rset=user.cw_rset.copy(), row=user.cw_row, col=user.cw_col)
    clone.cw_attr_cache = dict(user.cw_attr_cache)
    return clone


def cnx_attach_entity(cnx, entity):
    """Attach an entity to a cnx."""
    entity._cw = cnx
    if entity.cw_rset:
        entity.cw_rset.req = cnx


@lru_cache(10)
def cached_build_user(repo, eid):
    """Cached version of
    :meth:`cubicweb.server.repository.Repository._build_user`
    """
    with repo.internal_cnx() as cnx:
        user = repo._build_user(cnx, eid)
        lang = user.prefered_language()
        user.cw_clear_relation_cache()
        return clone_user(repo, user), lang



from functools import partial

from cubicweb.schema import PURE_VIRTUAL_RTYPES
from cubicweb.server.schema2sql import build_index_name, check_constraint

sql = partial(sql, ask_confirm=False)

source = repo.system_source
helper = source.dbhelper

# drop all relations primary keys
for table, cstr in sql(
    r"""
    SELECT DISTINCT tc.table_name, tc.constraint_name
    FROM information_schema.table_constraints tc,
         information_schema.key_column_usage kc
    WHERE tc.constraint_type = 'PRIMARY KEY'
          AND kc.table_name = tc.table_name
          AND kc.table_name LIKE '%\_relation'
          AND kc.table_schema = tc.table_schema
          AND kc.constraint_name = tc.constraint_name;
"""
):
    sql(f"ALTER TABLE {table} DROP CONSTRAINT {cstr}")

for table, cstr in sql(
    r"""
    SELECT DISTINCT table_name, constraint_name FROM information_schema.constraint_column_usage
    WHERE table_name LIKE 'cw\_%' AND constraint_name LIKE '%\_key'"""
):
    sql(f"ALTER TABLE {table} DROP CONSTRAINT {cstr}")

for rschema in schema.relations():
    if rschema.rule or rschema in PURE_VIRTUAL_RTYPES:
        continue
    if rschema.final or rschema.inlined:
        for rdef in rschema.rdefs.values():
            table = f"cw_{rdef.subject}"
            column = f"cw_{rdef.rtype}"
            if rschema.inlined or rdef.indexed:
                old_name = f"{table.lower()}_{column.lower()}_idx"
                sql(f"DROP INDEX IF EXISTS {old_name}")
                source.create_index(cnx, table, column)
    else:
        table = f"{rschema}_relation"
        sql(
            "ALTER TABLE %s ADD CONSTRAINT %s PRIMARY KEY(eid_from, eid_to)"
            % (table, build_index_name(table, ["eid_from", "eid_to"], "key_"))
        )
        for column in ("from", "to"):
            sql(f"DROP INDEX IF EXISTS {table}_{column}_idx")
            sql(
                "CREATE INDEX %s ON %s(eid_%s);"
                % (build_index_name(table, ["eid_" + column], "idx_"), table, column)
            )

# we changed constraint serialization, which also changes their name

for table, cstr in sql(
    """
    SELECT DISTINCT table_name, constraint_name FROM information_schema.constraint_column_usage
    WHERE constraint_name LIKE 'cstr%'"""
):
    sql(f"ALTER TABLE {locals()['table']} DROP CONSTRAINT {locals()['cstr']}")

for cwconstraint in rql("Any C WHERE R constrained_by C").entities():
    cwrdef = cwconstraint.reverse_constrained_by[0]
    rdef = cwrdef.yams_schema()
    cstr = rdef.constraint_by_eid(cwconstraint.eid)
    with cnx.deny_all_hooks_but():
        cwconstraint.cw_set(value=str(cstr.serialize()))
    if cstr.type() not in (
        "BoundaryConstraint",
        "IntervalBoundConstraint",
        "StaticVocabularyConstraint",
    ):
        # These cannot be translate into backend CHECK.
        continue
    cstrname, check = check_constraint(rdef, cstr, helper, prefix="cw_")
    args = {"e": rdef.subject.type, "c": cstrname, "v": check}
    sql(f"ALTER TABLE cw_{args['e']} ADD CONSTRAINT {args['c']} CHECK({args['v']})")

commit()

if "identity_relation" in helper.list_tables(cnx.cnxset.cu):
    sql("DROP TABLE identity_relation")
